<?php
/**
 * @package: SobiPro Library
 *
 * @author
 * Name: Sigrid Suski & Radek Suski, Sigsiu.NET GmbH
 * Email: sobi[at]sigsiu.net
 * Url: https://www.Sigsiu.NET
 *
 * @copyright Copyright (C) 2006 - 2019 Sigsiu.NET GmbH (https://www.sigsiu.net). All rights reserved.
 * @license GNU/LGPL Version 3
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License version 3
 * as published by the Free Software Foundation, and under the additional terms according section 7 of GPL v3.
 * See https://www.gnu.org/licenses/lgpl.html and https://www.sigsiu.net/licenses.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */

defined( 'SOBIPRO' ) || exit( 'Restricted access' );

SPLoader::loadView( 'section' );

/**
 * @author Radek Suski
 * @created 29-March-2010 13:08:21
 * @modified 25 September 2019 by Sigrid Suski
 */
class SPSearchView extends SPSectionView implements SPView
{
	/**
	 * @param string $type
	 * @param null $out
	 *
	 * @throws SPException
	 * @throws \Sobi\Error\Exception
	 */
	public function display( $type = 'section', $out = null )
	{
		$this->_type = 'search';
		$type = $this->key( 'template_type', 'xslt' );
		if ( $type != 'php' && Sobi::Cfg( 'global.disable_xslt', false ) ) {
			$type = 'php';
		}
		if ( $type == 'xslt' ) {
			$searchData = [];
			$fields = $this->get( 'fields' );
			$visitor = $this->get( 'visitor' );
			$p = $this->get( 'priorities' );
			$priorities = [];
			if ( is_array( $p ) && count( $p ) ) {
				foreach ( $p as $priority => $eids ) {
					if ( is_array( $eids ) && count( $eids ) ) {
						foreach ( $eids as $sid ) {
							$priorities[ $sid ] = $priority;
						}
					}
				}
			}
			$entries = $this->get( 'entries' );
			$searchData[ 'section' ] = [
				'_complex'    => 1,
				'_data'       => Sobi::Section( true ),
				'_attributes' => [ 'id' => Sobi::Section(), 'lang' => Sobi::Lang( false ) ],
			];
			$section = SPFactory::Section( Sobi::Section() );
			$searchData[ 'name' ] = [
				'_complex'    => 1,
				'_data'       => $section->get( 'sfTitle' ),
				'_attributes' => [ 'lang' => Sobi::Lang( false ) ],
			];
			$css_debug = '';
			if ( $development = ( Sobi::Cfg( 'template.development', true ) && !defined( 'SOBIPRO_ADM' ) ) ) {
				$searchData[ 'development' ] = $development;
				$css_debug = ' development';
			}
			$searchData[ 'description' ] = [
				'_complex'    => 1,
				'_data'       => $section->get( 'sfDesc' ),
				'_attributes' => [ 'lang' => Sobi::Lang( false ) ],
			];

			$searchPhrase = $this->get( 'search_for' );
			$phrase = $this->get( 'search_phrase' );
			$searchPhrase = strlen( $searchPhrase ) ? $searchPhrase : Sobi::Txt( 'SH.SEARCH_FOR_BOX' );
/* new for radius search ---------------------------- */
            $loadradius = Sobi::Cfg( 'search.radius', true ); // enter false to disable radius-search without updated general.xlm
            $unit =  (Sobi::Cfg( 'radius.unit', 1 )) ? 'km' : 'mi'; // 1 =km 0 =miles

            $searchDist = $this->get( 'radius_select' );
            $searchCity = $this->get( 'radius_city' );
            $searchCity = strlen( $searchCity ) ? ucfirst($searchCity) : Sobi::Txt( 'SH.SEARCH_FOR_DIST' );
/* end for radius search ---------------------------- */

            SPFactory::header()->addJsCode( 'var spSearchDefStr = "' . Sobi::Txt( 'SH.SEARCH_FOR_BOX' ) . '"' );
//No NEED   SPFactory::header()->addJsCode( 'var spSearchRadStr = "' . Sobi::Txt( 'SH.SEARCH_FOR_DIST' ) . '";' );

            if ( $this->get( '$eInLine' ) ) {
				$searchData[ 'entries_in_line' ] = $this->get( '$eInLine' );
			}
			$scount = $this->get( '$eCount' );
			$searchData[ 'count' ] = $scount;
			if ( $scount >= 0 ) {
			    $m = ($this->get( '$eCount' ) <> 1) ? Sobi::Txt('SH.MULTICHAR') : '';
                $searchWord = ($searchPhrase == Sobi::Txt( 'SH.SEARCH_FOR_BOX' )) ? Sobi::Txt( 'SH.SEARCH_FILTER' ) : $searchPhrase;
                $sterm = $this->get( 'search_for' );
/*result message*/ if ( $this->get( 'searchtype' ) == 'radius' ) {
/*for radius and*/ if ($this->get( 'error' ) == 'geo')
/*error added */   $searchData[ 'message' ] = 'ERROR. '. Sobi::Txt( 'SH.SEARCH_HASH_ERROR', ['id' => $this->get('radius_city')] );
                   elseif ($this->get( 'error' ) == 'key')
                   $searchData[ 'message' ] = 'ERROR. '. Sobi::Txt( 'SH.SEARCH_API_ERROR' );
                   else
                   $searchData[ 'message' ] = Sobi::Txt('SH.SEARCH_FOUND_DISTRESULTS', [ 'multi' => $m, 'count' => $scount, 'city' => $searchCity,'dist' => $searchDist]) ;
                }
				elseif ( isset( $sterm ) && ( $sterm != '*' ) && ( $sterm != Sobi::Txt( 'SH.SEARCH_FOR_BOX' ) ) ) {
					$searchData[ 'message' ] = Sobi::Txt( 'SH.SEARCH_FOUND_TERM_RESULTS', [ 'multi' => $m, 'count' => $scount, 'sterm' => $sterm ] );
				}
				elseif ( isset( $sterm ) && $sterm == '*' ) {
					$searchData[ 'message' ] = Sobi::Txt( 'SH.SEARCH_FOUND_ALL', [ 'multi' => $m, 'count' => $scount ] );
				}
				else {
					$searchData[ 'message' ] = Sobi::Txt( 'SH.SEARCH_FOUND_RESULTS', [ 'multi' => $m, 'count' => $scount, 'word' => $searchWord ] );
				}
			}
			$searchData[ 'search_order' ] = $this->get( 'orderings' );
			$this->menu( $searchData );
			$this->alphaMenu( $searchData );
			$fData = [];
			if ( Sobi::Cfg( 'search.show_searchbox', true ) ) {
				$fData[ 'searchbox' ] = [
					'_complex'    => 1,
					'_data'       => [
						'label' => [
							'_complex'    => 1,
							'_data'       => Sobi::Txt( 'SH.SEARCH_FOR' ),
							'_attributes' => [ 'lang' => Sobi::Lang( false ) ],
						],
						'data'  => [
							'_complex' => 1,
							'_xml'     => 1,
							'_data'    => SPHtml_Input::text( 'sp_search_for', $searchPhrase, [ 'class' => Sobi::Cfg( 'search.form_box_def_css', 'SPSearchBox' ), 'id' => 'SPSearchBox' ] ),
						],
					],
					'_attributes' => [ 'position' => 1, 'css_class' => 'SPSearchBox' ],
				];
			}

			if ( Sobi::Cfg( 'search.top_button', true ) ) {
				$fData[ 'top_button' ] = [
					'_complex'    => 1,
					'_data'       => [
						'label' => [
							'_complex'    => 1,
							'_data'       => Sobi::Txt( 'SH.SEARCH_START' ),
							'_attributes' => [ 'lang' => Sobi::Lang() ],
						],
						'data'  => [
							'_complex' => 1,
							'_xml'     => 1,
							'_data'    => SPHtml_Input::submit( 'search', Sobi::Txt( 'SH.START' ), [ 'id' => 'top_button' ] ),
						],
					],
					'_attributes' => [ 'position' => 1, 'css_class' => 'SPSearchButton' ],
				];
			}

			if ( Sobi::Cfg( 'search.show_phrase', true ) ) {
				$fData[ 'phrase' ] = [
					'_complex'    => 1,
					'_data'       => [
						'label' => [
							'_complex'    => 1,
							'_data'       => Sobi::Txt( 'SH.FIND_ENTRIES_THAT_HAVE' ),
							'_attributes' => [ 'lang' => Sobi::Lang( false ) ],
						],
						'data'  => [
							'_complex' => 1,
							'_xml'     => 1,
							'_data'    => SPHtml_Input::radioList(
								'spsearchphrase',
								[
									'all'   => Sobi::Txt( 'SH.FIND_ENTRIES_THAT_HAVE_ALL_WORDS' ),
									'any'   => Sobi::Txt( 'SH.FIND_ENTRIES_THAT_HAVE_ANY_WORDS' ),
									'exact' => Sobi::Txt( 'SH.FIND_ENTRIES_THAT_HAVE_EXACT_PHRASE' ),
								],
								'spsearchphrase',
								strlen( $phrase ) ? $phrase : Sobi::Cfg( 'search.form_searchphrase_def', 'all' ),
								null,
								'right'
							),
						],
					],
					'_attributes' => [ 'position' => 1, 'css_class' => 'SPSearchPhrase' ],
				];
			}
/* ------- this is for the new radius-search container ------------------- */
            if ( $loadradius ) {
                        $fData[ 'radiusbox' ] = [
            					'_complex'    => 1,
            					'_data'       => [
            						'label' => [
            							'_complex'    => 1,
            							'_data'       => Sobi::Txt( 'SH.SEARCH_DIST' ),
            							'_attributes' => [ 'lang' => Sobi::Lang( false ) ],
            						],
            						'data'  => [
                                    	'_complex' => 1,
            							'_xml'     => 1,
                                    	'_data'    => SPHtml_Input::text( 'sp_radius_city', $searchCity, [ 'class' => Sobi::Cfg( 'search.form_box_def_css', 'SPSearchBox' ), 'id' => 'SPRadiusCity' ] ),
                                        ],
            					],
            					'_attributes' => [ 'position' => 1, 'css_class' => 'SPSearchBox' ],
            				];

                        $fData[ 'radiusdist' ] = [
            					'_complex'    => 1,
            					'_data'       => [
            						'label' => [
            							'_complex'    => 1,
            							'_data'       => Sobi::Txt( 'SH.FIND_ENTRIES_WITHIN' ),
            							'_attributes' => [ 'lang' => Sobi::Lang( false ) ],
            						],
            						'data'  => [
            							'_complex' => 1,
            							'_xml'     => 1,
            							'_data'    => SPHtml_Input::select(
            								'sp_radius_select',
            								[
            									5    => '&emsp;5 '.$unit,
            									10   => '&ensp;10 '.$unit,
            									25   => '&ensp;25 '.$unit,
                                                50   => '&ensp;50 '.$unit,
            									100  => '100 '.$unit,
            									150  => '150 '.$unit,
                                                200  => '200 '.$unit,
                                                250  => '250 '.$unit,
            								],
                                            $searchDist,
                                            false,
                                            null ),
            						],
            					],
            					'_attributes' => [ 'position' => 1, 'css_class' => 'SPField form-control inputbox' ],
            				];

                        $fData[ 'radius_button' ] = [
            					'_complex'    => 1,
            					'_data'       => [
            						'label' => [
            							'_complex'    => 1,
            							'_data'       => Sobi::Txt( 'SH.SEARCH_START' ),
            							'_attributes' => [ 'lang' => Sobi::Lang( false ) ],
            						],
            						'data'  => [
            							'_complex' => 1,
            							'_xml'     => 1,
            							'_data'    => SPHtml_Input::submit( 'search', Sobi::Txt( 'SH.RADSTART' ), [ 'id' => 'rad_button' ] ),
            						],
            					],
            					'_attributes' => [ 'position' => 1, 'css_class' => 'SPSearchButton' ],
            				];
            }
/* ------- END new radius-search container ------------------- */
			if ( count( $fields ) ) {
				foreach ( $fields as $field ) {
					$data = $field->searchForm();
					$suffix = $field->get( 'searchMethod' ) != 'range' ? $field->get( 'suffix' ) : null;
					if ( strlen( $data ) ) {
						$fData[ $field->get( 'nid' ) ] = [
							'_complex'    => 1,
							'_data'       => [
								'label' => [
									'_complex'    => 1,
									'_data'       => $field->get( 'name' ),
									'_attributes' => [ 'lang' => Sobi::Lang() ],
								],
								'data'  => [
									'_complex' => 1,
									'_xml'     => 1,
									'_data'    => $data,
								],
							],
							'_attributes' => [ 'id'         => $field->get( 'id' ),
							                   'type'       => $field->get( 'type' ),
							                   'suffix'     => $suffix,
							                   'position'   => $field->get( 'position' ),
							                   'css_search' => $field->get( 'cssClassSearch' ) . $css_debug,
							                   'debug'      => $development,
							                   'width'      => $field->get( 'bsSearchWidth' ),
							                   'css_class'  => ( strlen( $field->get( 'cssClass' ) ) ? $field->get( 'cssClass' ) : 'spField' ),
							],
						];
// not needed for radius search - it enables the section (with value "") in cat. dropdown, so it can be used as deselect option or use javascript with id of category_field
                    if ($field->get( 'type' ) == 'category') {
                    $fData[ $field->get( 'nid' ) ]['_data']['data']['_data'] = preg_replace('/disabled="disabled" value="\d*/','value="',$data);
                    }
                // end of this new block 
					}
				}
			}
			if ( Sobi::Cfg( 'search.bottom_button', false ) ) {
				$fData[ 'bottom_button' ] = [
					'_complex'    => 1,
					'_data'       => [
						'label' => [
							'_complex'    => 1,
							'_data'       => Sobi::Txt( 'SH.SEARCH_START' ),
							'_attributes' => [ 'lang' => Sobi::Lang( false ) ],
						],
						'data'  => [
							'_complex' => 1,
							'_xml'     => 1,
							'_data'    => SPHtml_Input::submit( 'search', Sobi::Txt( 'SH.START' ) ),
						],
					],
					'_attributes' => [ 'position' => 1, 'css_class' => 'SPSearchButton' ],
				];
			}

			$searchData[ 'fields' ] = $fData;

/*new 1.5*/ $this->switchOrderingMenu( $searchData, $this->get( 'orderings' ), Sobi::Cfg( 'sordering.fields_array' ) );

			if ( is_array( $entries ) && count( $entries ) ) {
				$this->loadNonStaticData( $entries );
				$manager = Sobi::Can( 'entry', 'edit', '*', Sobi::Section() ) ? true : false;
				foreach ( $entries as $entry ) {
					$en = $this->entry( $entry, $manager );
					$searchData[ 'entries' ][] = [
						'_complex'    => 1,
/*!!edit for radius*/   '_attributes' => [ 'id' => $en[ 'id' ], 'km' => $this->get('dist_val')[$en[ 'id' ]], 'unit' => $unit, 'search-priority' => isset( $priorities[ $en[ 'id' ] ] ) ? $priorities[ $en[ 'id' ] ] : 'undefined' ],
						'_data'       => $en,                  //IMPORTANT: this adds distance and unit to the vcards !!!!
					];
				}
				$this->navigation( $searchData );
			}
			$searchData[ 'visitor' ] = $this->visitorArray( $visitor );
			$this->_attr = $searchData;
		}
		SPFactory::config()->nameField();   // Check for status of name field and output a warning if field does not exist or is disabled
		Sobi::Trigger( $this->_type, ucfirst( __FUNCTION__ ), [ &$this->_attr ] );
		parent::display( $this->_type );
	}
}
